from click.testing import CliRunner
from io import StringIO
import pytest
import re
import subprocess


@pytest.yield_fixture(scope="session", autouse=True)
def event_loop_closing():
    import check_tls_certs
    loop = set()

    def close_event_loop(_loop):
        loop.add(_loop)

    check_tls_certs.close_event_loop = close_event_loop
    yield
    for l in loop:
        l.close()


@pytest.fixture
def no_chain_validation(monkeypatch):
    monkeypatch.setattr(
        "check_tls_certs.validate_certificate_chain",
        lambda c, m: None)


@pytest.fixture
def makecert():
    import OpenSSL

    def makecert(cn='example.com', days=365):
        result = subprocess.check_output([
            'openssl',
            'req', '-x509',
            '-nodes',
            '-newkey', 'rsa:1024',
            '-days', '%s' % days,
            '-subj', '/C=DE/ST=Foo/L=Bar/O=Foo LTD/OU=Org/CN=%s' % cn])
        m = re.search(
            b'(-----BEGIN CERTIFICATE-----.*?-----END CERTIFICATE-----)',
            result,
            flags=re.DOTALL)
        return OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM,
            m.group(1))

    return makecert


@pytest.fixture
def utcnow():
    import datetime
    return datetime.datetime.utcnow() + datetime.timedelta(seconds=1)


def test_arg(monkeypatch):
    from check_tls_certs import main
    domain_fetches = []
    monkeypatch.setattr(
        "check_tls_certs.get_cert_from_domain",
        lambda x: domain_fetches.append(x) or (x, None))
    domain_checks = []
    monkeypatch.setattr(
        "check_tls_certs.check",
        lambda x, u: domain_checks.append(x) or ([], None))
    runner = CliRunner()
    result = runner.invoke(main, ['-vv', 'example.com/www.example.com'])
    assert result.exit_code == 0
    assert [[d for d, c in x] for x in domain_checks] == [
        ['example.com', 'www.example.com']]
    assert result.output == (
        'example.com, www.example.com\n'
        '0 error(s), 0 warning(s)\n')


def test_file(monkeypatch, tmpdir):
    from check_tls_certs import main
    domain_fetches = []
    monkeypatch.setattr(
        "check_tls_certs.get_cert_from_domain",
        lambda x: domain_fetches.append(x) or (x, None))
    domain_checks = []
    monkeypatch.setattr(
        "check_tls_certs.check",
        lambda x, u: domain_checks.append(x) or ([], None))
    f = tmpdir.join("domains.txt")
    f.write_binary(b"example.com/www.example.com\nfoo.com\nbar.com/\n    #none.bar.com\n    www.bar.com")
    runner = CliRunner()
    result = runner.invoke(main, ['-vv', '-f', f.strpath])
    assert result.exit_code == 0
    assert [[d for d, c in x] for x in domain_checks] == [
        ['example.com', 'www.example.com'],
        ['foo.com'],
        ['bar.com', 'www.bar.com']]
    assert result.output == (
        'example.com, www.example.com\n'
        'foo.com\n'
        'bar.com, www.bar.com\n'
        '0 error(s), 0 warning(s)\n')


def test_domain_definitions_from_cli_parse_error(capsys):
    from check_tls_certs import domain_definitions_from_cli
    with pytest.raises(SystemExit) as e:
        domain_definitions_from_cli(["foo:bar"])
    (out, err) = capsys.readouterr()
    assert "Error in definition 'foo:bar': Couldn't parse 'foo:bar', port 'bar' is not an integer" in err
    assert e.value.args == (5,)


def test_domain_definitions_from_lines_parse_error(capsys):
    from check_tls_certs import domain_definitions_from_lines
    with pytest.raises(SystemExit) as e:
        domain_definitions_from_lines(StringIO("foo:bar"))
    (out, err) = capsys.readouterr()
    assert "Error in definition starting on line 1: Couldn't parse 'foo:bar', port 'bar' is not an integer" in err
    assert e.value.args == (5,)
    with pytest.raises(SystemExit) as e:
        domain_definitions_from_lines(StringIO("# comment\nfoo:bar"))
    (out, err) = capsys.readouterr()
    assert "Error in definition starting on line 2: Couldn't parse 'foo:bar', port 'bar' is not an integer" in err
    assert e.value.args == (5,)
    with pytest.raises(SystemExit) as e:
        domain_definitions_from_lines(StringIO("# comment\nfoo:2/ham:egg"))
    (out, err) = capsys.readouterr()
    assert "Error in definition starting on line 2: Couldn't parse 'ham:egg', port 'egg' is not an integer" in err
    assert e.value.args == (5,)
    with pytest.raises(SystemExit) as e:
        domain_definitions_from_lines(StringIO("# comment\nfoo:2\nham:egg"))
    (out, err) = capsys.readouterr()
    assert "Error in definition starting on line 3: Couldn't parse 'ham:egg', port 'egg' is not an integer" in err
    assert e.value.args == (5,)


def test_domain_no_fetch():
    from check_tls_certs import Domain
    d = Domain('!foo')
    assert d.no_fetch
    assert d.host == 'foo'


def test_domain_connection_host():
    from check_tls_certs import Domain
    d = Domain('foo')
    assert d.host == 'foo'
    assert d.connection_host == 'foo'
    assert d == 'foo'
    d = Domain('foo|bar')
    assert d.host == 'foo'
    assert d.connection_host == 'bar'
    assert d == 'foo (bar)'


def test_get_cert_from_domain_no_fetch():
    from check_tls_certs import Domain
    from check_tls_certs import get_cert_from_domain
    d = Domain('!foo')
    assert get_cert_from_domain(d) == (d, None)
    assert get_cert_from_domain(d) == ("foo", None)


def test_get_cert_from_domain_socket_gaierror(monkeypatch):
    from check_tls_certs import Domain
    from check_tls_certs import get_cert_from_domain
    from unittest import mock
    import socket
    _get_cert_from_domain = mock.Mock()
    _get_cert_from_domain.side_effect = socket.gaierror
    monkeypatch.setattr(
        "check_tls_certs._get_cert_from_domain",
        _get_cert_from_domain)
    with pytest.raises(socket.gaierror):
        get_cert_from_domain(Domain('foo'))


def test_get_cert_from_domain_other_error(monkeypatch):
    from check_tls_certs import Domain
    from check_tls_certs import get_cert_from_domain
    from unittest import mock
    _get_cert_from_domain = mock.Mock()
    _get_cert_from_domain.side_effect = ValueError("ham")
    monkeypatch.setattr(
        "check_tls_certs._get_cert_from_domain",
        _get_cert_from_domain)
    d = Domain('foo')
    result = get_cert_from_domain(d)
    assert result == (d, "ValueError: ham")


def test_check_self_signed(makecert, utcnow):
    from check_tls_certs import Domain
    from check_tls_certs import check
    d = Domain('example.com')
    cert = makecert()
    (msgs, earliest_expiration) = check([(d, [cert])], utcnow)
    (errmsg,) = [m for m in msgs if m[0] == 'error']
    assert 'Validation error' in errmsg[1]
    assert 'self signed certificate' in errmsg[1]


@pytest.mark.usefixtures('no_chain_validation')
def test_expiration_far_in_future(makecert, utcnow):
    from check_tls_certs import Domain
    from check_tls_certs import check
    d = Domain('example.com')
    cert = makecert()
    (msgs, earliest_expiration) = check([(d, [cert])], utcnow)
    (msg,) = [m for m in msgs if m[1].startswith('Valid until')]
    assert '(364 days,' in msg[1]
    assert (earliest_expiration - utcnow).days == 364


@pytest.mark.usefixtures('no_chain_validation')
def test_expiration_within_warning_range(makecert, utcnow):
    from check_tls_certs import Domain
    from check_tls_certs import check
    from check_tls_certs import default_expiry_warn
    d = Domain('example.com')
    cert = makecert(days=default_expiry_warn)
    (msgs, earliest_expiration) = check([(d, [cert])], utcnow)
    (msg,) = [m for m in msgs if 'certificate expires on' in m[1]]
    assert '(%s days,' % (default_expiry_warn - 1) in msg[1]
    assert (earliest_expiration - utcnow).days == (default_expiry_warn - 1)


@pytest.mark.usefixtures('no_chain_validation')
def test_expiration_expired(makecert, utcnow):
    from check_tls_certs import Domain
    from check_tls_certs import check
    import datetime
    d = Domain('example.com')
    cert = makecert(days=1)
    utcnow = utcnow + datetime.timedelta(days=1)
    (msgs, earliest_expiration) = check([(d, [cert])], utcnow)
    (msg,) = [m for m in msgs if 'certificate has expired on' in m[1]]
    assert msg[1].startswith('The certificate has expired on')
    assert (earliest_expiration - utcnow).days < 0


@pytest.mark.usefixtures('no_chain_validation')
def test_wildcard(makecert, utcnow):
    from check_tls_certs import Domain
    from check_tls_certs import check
    cert = makecert(cn='*.example.com')
    d = Domain('www.example.com')
    (msgs, earliest_expiration) = check([(d, [cert])], utcnow)
    errs = [m for m in msgs if "doesn't match the certificate domain" in m[1]]
    assert errs == []
    assert (earliest_expiration - utcnow).days > 360
    d = Domain('foo.bar.example.com')
    (msgs, earliest_expiration) = check([(d, [cert])], utcnow)
    (err,) = [m for m in msgs if "doesn't match the certificate domain" in m[1]]
    assert "foo.bar.example.com doesn't match the certificate domain *.example.com" in err[1]
    assert (earliest_expiration - utcnow).days > 360



import asyncio
import click
import datetime
import itertools
import select
import socket
import sys
import traceback
import OpenSSL


default_expiry_warn = 14


class DomainParseError(Exception):
    pass


class Domain(str):
    def __new__(cls, domain):
        if domain.startswith('!'):
            name = domain[1:]
        else:
            name = domain
        host = name
        port = 443
        if ':' in name:
            host, port = name.split(':')
            try:
                port = int(port)
            except ValueError:
                raise DomainParseError("Couldn't parse '%s', port '%s' is not an integer" % (domain, port))
        connection_host = host
        if '|' in name:
            host, connection_host = name.split('|')
            name = "%s (%s)" % (host, connection_host)
        result = str.__new__(cls, name)
        if domain.startswith('!'):
            result.no_fetch = True
        else:
            result.no_fetch = False
        result.host = host
        result.connection_host = connection_host
        result.port = port
        return result


class CertDomains(list):
    def __init__(self, domain_definitions):
        for d in domain_definitions.split('/'):
            self.append(Domain(d))


def fatal(msg):
    click.echo(click.style(msg, fg="red"), err=True)
    sys.exit(5)


def _get_cert_from_domain(domain):
    ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
    sock = OpenSSL.SSL.Connection(ctx, socket.socket())
    sock.settimeout(5)
    sock.set_tlsext_host_name(domain.encode('ascii'))
    sock.connect((domain.connection_host, domain.port))
    while True:
        try:
            sock.do_handshake()
            break
        except OpenSSL.SSL.WantReadError:
            select.select([sock], [], [])
    return sock.get_peer_cert_chain()


def get_cert_from_domain(domain):
    if domain.no_fetch:
        return (domain, None)
    try:
        data = _get_cert_from_domain(domain)
    except socket.gaierror:
        raise
    except Exception as e:
        data = "".join(traceback.format_exception_only(type(e), e)).strip()
    return (domain, data)


def close_event_loop(loop):
    # just a separate function, so this can be monkeypatched in tests
    loop.close()


def get_domain_certs(domains):
    loop = asyncio.get_event_loop()
    (done, pending) = loop.run_until_complete(asyncio.wait([
        loop.run_in_executor(None, get_cert_from_domain, x)
        for x in itertools.chain(*domains)]))
    close_event_loop(loop)
    return dict(x.result() for x in done)


def domain_key(d):
    return tuple(reversed(d.split('.')))


def validate_certificate_chain(cert_chain, msgs):
    ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
    ctx.set_default_verify_paths()
    cert_store = ctx.get_cert_store()
    for index, cert in reversed(list(enumerate(cert_chain))):
        sc = OpenSSL.crypto.X509StoreContext(cert_store, cert)
        try:
            sc.verify_certificate()
        except OpenSSL.crypto.X509StoreContextError as e:
            msgs.append(
                ('error', "Validation error '%s'." % e))
        if index > 0:
            cert_store.add_cert(cert)


def check(domainnames_certs, utcnow, expiry_warn=default_expiry_warn):
    msgs = []
    domainnames = set(dnc[0].host for dnc in domainnames_certs)
    earliest_expiration = None
    for domain, cert_chain in domainnames_certs:
        if cert_chain is None:
            continue
        if not any(isinstance(cert, OpenSSL.crypto.X509) for cert in cert_chain):
            msgs.append(
                ('error', "Couldn't fetch certificate for %s.\n%s" % (domain, cert_chain)))
            continue
        validate_certificate_chain(cert_chain, msgs)
        cert = cert_chain[0]
        expires = datetime.datetime.strptime(cert.get_notAfter().decode('ascii'), '%Y%m%d%H%M%SZ')
        if expires:
            if earliest_expiration is None or expires < earliest_expiration:
                earliest_expiration = expires
        issued_level = "info"
        issuer = cert.get_issuer().commonName
        if issuer.lower() == "happy hacker fake ca":
            issued_level = "error"
        msgs.append(
            (issued_level, "Issued by: %s" % issuer))
        if len(cert_chain) > 1:
            sign_cert = cert_chain[1]
            subject = sign_cert.get_subject().commonName
            if issuer != subject:
                msgs.append(
                    ('error', "The certificate sign chain subject '%s' doesn't match the issuer '%s'." % (subject, issuer)))
        sig_alg = cert.get_signature_algorithm()
        if sig_alg.startswith(b'sha1'):
            msgs.append(
                ('error', "Unsecure signature algorithm %s" % sig_alg))
        if expires < utcnow:
            msgs.append(
                ('error', "The certificate has expired on %s." % expires))
        elif expires < (utcnow + datetime.timedelta(days=expiry_warn)):
            msgs.append(
                ('warning', "The certificate expires on %s (%s)." % (
                    expires, expires - utcnow)))
        else:
            # rounded delta
            delta = ((expires - utcnow) // 60 // 10 ** 6) * 60 * 10 ** 6
            msgs.append(
                ('info', "Valid until %s (%s)." % (expires, delta)))
        alt_names = set()
        for index in range(cert.get_extension_count()):
            ext = cert.get_extension(index)
            if ext.get_short_name() != b'subjectAltName':
                continue
            alt_names.update(
                x.strip().replace('DNS:', '')
                for x in str(ext).split(','))
        alt_names.add(cert.get_subject().commonName)
        unmatched = domainnames.difference(alt_names)
        if unmatched:
            msgs.append(
                ('info', "Alternate names in certificate: %s" % ', '.join(
                    sorted(alt_names, key=domain_key))))
            if len(domainnames) == 1:
                name = cert.get_subject().commonName
                if name != domain.host:
                    if name.startswith('*.'):
                        name_parts = name.split('.')[1:]
                        name_parts_len = len(name_parts)
                        domain_host_parts = domain.host.split('.')
                        if (len(domain_host_parts) - name_parts_len) == 1:
                            if domain_host_parts[-name_parts_len:] == name_parts:
                                continue
                    msgs.append(
                        ('error', "The requested domain %s doesn't match the certificate domain %s." % (domain, name)))
            else:
                msgs.append(
                    ('warning', "Unmatched alternate names %s." % ', '.join(
                        sorted(unmatched, key=domain_key))))
        elif domainnames == alt_names:
            msgs.append(
                ('info', "Alternate names match specified domains."))
        else:
            unmatched = alt_names.difference(domainnames)
            msgs.append(
                ('warning', "More alternate names than specified %s." % ', '.join(
                    sorted(unmatched, key=domain_key))))
    return (msgs, earliest_expiration)


def check_domains(domains, domain_certs, utcnow):
    result = []
    for domainnames in domains:
        domainnames_certs = [(dn, domain_certs[dn]) for dn in domainnames]
        msgs = []
        seen = set()
        earliest_expiration = None
        (dmsgs, expiration) = check(domainnames_certs, utcnow)
        for level, msg in dmsgs:
            if expiration:
                if earliest_expiration is None or expiration < earliest_expiration:
                    earliest_expiration = expiration
            if msg not in seen:
                seen.add(msg)
                msgs.append((level, msg))
        result.append((domainnames, msgs, earliest_expiration))
    return result


def domain_definitions_from_lines(lines):
    result = []
    current = []
    current_start = 1
    for index, line in enumerate(lines, start=1):
        line = line.strip()
        if not line or line.startswith('#'):
            if not current:
                current_start += 1
            continue
        if line.endswith('/'):
            current.append(line)
            continue
        current.append(line)
        try:
            domain_definition = CertDomains("".join(current))
        except DomainParseError as e:
            fatal("Error in definition starting on line %s: %s" % (current_start, e))
        result.append(domain_definition)
        current = []
        current_start = index + 1
    return result


def domain_definitions_from_filename(filename):
    if not filename:
        return []
    with open(filename, 'r', encoding='utf-8') as f:
        return domain_definitions_from_lines(f)


def domain_definitions_from_cli(domains):
    result = []
    if not domains:
        return result
    for domain in domains:
        try:
            domain_definition = CertDomains(domain)
        except DomainParseError as e:
            fatal("Error in definition '%s': %s" % (domain, e))
        result.append(domain_definition)
    return result


@click.command()
@click.option('-f', '--file', metavar='FILE', help='File to read domains from. One per line.')
@click.option('-v', '--verbose', count=True, help='Increase verbosity. Can be used several times. Currently max verbosity is 2.')
@click.argument('domain', nargs=-1)
def main(file, domain, verbose):
    """Checks the TLS certificate for each DOMAIN.

       You can add checks for alternative names by separating them with a slash, like example.com/www.example.com.

       Wildcard domains are supported.

       Exits with return code 3 when there are warnings, code 4 when there are errors and code 5 when the domain definition contains errors.
    """
    domains = list(itertools.chain(
        domain_definitions_from_filename(file),
        domain_definitions_from_cli(domain)))
    domain_certs = get_domain_certs(domains)
    total_warnings = 0
    total_errors = 0
    earliest_expiration = None
    utcnow = datetime.datetime.utcnow()
    for domainnames, msgs, expiration in check_domains(domains, domain_certs, utcnow):
        if expiration:
            if earliest_expiration is None or expiration < earliest_expiration:
                earliest_expiration = expiration
        warnings = 0
        errors = 0
        domain_msgs = [', '.join(domainnames)]
        for level, msg in msgs:
            if level == 'error':
                color = 'red'
                errors = errors + 1
            elif level == 'warning':
                color = 'yellow'
                warnings = warnings + 1
            else:
                color = None
            if color:
                msg = click.style(msg, fg=color)
            msg = "\n".join("    " + m for m in msg.split('\n'))
            domain_msgs.append(msg)
        if (verbose > 1) or warnings or errors:
            click.echo('\n'.join(domain_msgs))
        total_errors = total_errors + errors
        total_warnings = total_warnings + warnings
    msg = "%s error(s), %s warning(s)" % (total_errors, total_warnings)
    if earliest_expiration:
        msg += "\nEarliest expiration on %s (%s)." % (
            earliest_expiration, earliest_expiration - utcnow)
    if total_errors:
        click.echo(click.style(msg, fg="red"))
        sys.exit(4)
    elif total_warnings:
        click.echo(click.style(msg, fg="yellow"))
        sys.exit(3)
    if verbose:
        click.echo(click.style(msg, fg="green"))
